// FREAKOUT.CPP - break game demo

// INCLUDES ///////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN // include all macros
#define INITGUID            // include all GUIDs 

#include <windows.h>        // include important windows stuff
#include <windowsx.h> 
#include <mmsystem.h>

#include <iostream.h>       // include important C/C++ stuff
#include <conio.h>
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <math.h>
#include <io.h>
#include <fcntl.h>

#include <ddraw.h>          // directX includes
#include "blackbox.h"       // game library includes

#include "defines.h"

// DEFINES ////////////////////////////////////////////////////

// defines for windows 
#define WINDOW_CLASS_NAME "WIN3DCLASS"  // class name

#define WINDOW_WIDTH            640     // size of window
#define WINDOW_HEIGHT           480

// states for game loop
#define GAME_STATE_INIT         0
#define GAME_STATE_START_LEVEL  1
#define GAME_STATE_RUN          2
#define GAME_STATE_SHUTDOWN     3
#define GAME_STATE_EXIT         4 

// block defines
#define NUM_BLOCK_ROWS          6
#define NUM_BLOCK_COLUMNS       8

#define BLOCK_WIDTH             64
#define BLOCK_HEIGHT            16
#define BLOCK_ORIGIN_X          8
#define BLOCK_ORIGIN_Y          8
#define BLOCK_X_GAP             80
#define BLOCK_Y_GAP             32

// paddle defines
#define PADDLE_START_X          (SCREEN_WIDTH/2 - 16)
#define PADDLE_START_Y          (SCREEN_HEIGHT - 32);
#define PADDLE_WIDTH            32
#define PADDLE_HEIGHT           8
#define PADDLE_COLOR            191

// ball defines
#define BALL_COUNT				40
#define BALL_SIZE				4

// PROTOTYPES /////////////////////////////////////////////////

// game console
int Game_Init(void *parms=NULL);
int Game_Shutdown(void *parms=NULL);
int Game_Main(void *parms=NULL);

// GLOBALS ////////////////////////////////////////////////////

HWND main_window_handle  = NULL; // save the window handle
HINSTANCE main_instance  = NULL; // save the instance
int game_state           = GAME_STATE_INIT; // starting state

int paddle_x = 0, paddle_y = 0; // tracks position of paddle

struct Ball
{
	int x, y;
	int dx, dy;
} balls[BALL_COUNT];

int score    = 0;               // the score
int level    = 1;               // the current level
int blocks_hit = 0;             // tracks number of blocks hit

// this contains the game grid data   

UCHAR blocks[NUM_BLOCK_ROWS][NUM_BLOCK_COLUMNS];     

// FUNCTIONS //////////////////////////////////////////////////

LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	// this is the main message handler of the system
	PAINTSTRUCT	ps;		   // used in WM_PAINT
	HDC			hdc;	   // handle to a device context

	switch(msg)
		{	
		case WM_CREATE: 
			// do initialization stuff here
			return(0);
			break;

		case WM_PAINT:
			// start painting
			hdc = BeginPaint(hwnd,&ps);

			// the window is now validated 

			// end painting
			EndPaint(hwnd,&ps);
			return(0);
			break;

		case WM_DESTROY: 
			// kill the application			
			PostQuitMessage(0);
			return(0);
			break;

		default:
			break;

		} // end switch

	// process any messages that we didn't take care of 
	return (DefWindowProc(hwnd, msg, wparam, lparam));

} // end WinProc

// WINMAIN ////////////////////////////////////////////////////

int WINAPI WinMain(HINSTANCE hinstance, HINSTANCE hprevinstance, LPSTR lpcmdline, int ncmdshow)
{
	WNDCLASS winclass;	// this will hold the class we create
	HWND	 hwnd;		// generic window handle
	MSG		 msg;		// generic message
	HDC      hdc;       // generic dc
	PAINTSTRUCT ps;     // generic paintstruct

	// first fill in the window class stucture
	winclass.style			= CS_DBLCLKS | CS_OWNDC | 
							  CS_HREDRAW | CS_VREDRAW;
	winclass.lpfnWndProc	= WindowProc;
	winclass.cbClsExtra		= 0;
	winclass.cbWndExtra		= 0;
	winclass.hInstance		= hinstance;
	winclass.hIcon			= LoadIcon(NULL, IDI_APPLICATION);
	winclass.hCursor		= LoadCursor(NULL, IDC_ARROW);
	winclass.hbrBackground	= (HBRUSH)GetStockObject(BLACK_BRUSH);
	winclass.lpszMenuName	= NULL; 
	winclass.lpszClassName	= WINDOW_CLASS_NAME;

	// register the window class
	if (!RegisterClass(&winclass))
		return(0);

	// create the window, note the use of WS_POPUP
	if (!(hwnd = CreateWindow(WINDOW_CLASS_NAME,    // class
							  "WIN3D Game Console",	// title
							  WS_POPUP | WS_VISIBLE, 
							  0,0,	                // initial x,y
							  GetSystemMetrics(SM_CXSCREEN),  // intial width
							  GetSystemMetrics(SM_CYSCREEN),  // initial height
							  NULL,	    // handle to parent 
							  NULL,	    // handle to menu
							  hinstance,// instance
							  NULL)))	// creation parms
	return(0);

	ShowCursor(FALSE);	// hide mouse

	// save the window handle and instance in a global
	main_window_handle = hwnd;
	main_instance      = hinstance;

	Game_Init();		// perform all game console specific initialization

	// main event loop
	while(1)
		{
		if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))
			{ 
			// test if this is a quit
			if (msg.message == WM_QUIT)
			   break;
		
			TranslateMessage(&msg);	// translate any accelerator keys

			DispatchMessage(&msg);	// send the message to the window proc
			} // end if
    
		// main game processing goes here
		Game_Main();

		} // end while

	Game_Shutdown();	// shutdown game and release all resources

	ShowCursor(TRUE);	// show mouse

	// return to Windows like this
	return(msg.wParam);

} // end WinMain

// T3DX GAME PROGRAMMING CONSOLE FUNCTIONS ////////////////////

int Game_Init(void *parms)
{
	// this function is where you do all the initialization 
	// for your game

	DD_Init(SCREEN_WIDTH,SCREEN_HEIGHT,SCREEN_BPP);

	return(1);
} // end Game_Init

///////////////////////////////////////////////////////////////

int Game_Shutdown(void *parms)
{
	// this function is where you shutdown your game and
	// release all resources that you allocated

	DD_Shutdown();

	return(1);
} // end Game_Shutdown

inline bool CollisionTest(int px, int py, int r1left, int r1top, int r1right, int r1bottom)
{	return (px>=r1left) && (px<=r1right) && (py>=r1top) && (py<=r1bottom); }

inline bool CollisionTest(int px, int py, RECT rect)
{	return CollisionTest(px, py, rect.left, rect.top, rect.right, rect.bottom); }

inline bool CollisionTest(POINT pt, RECT rect)
{	return CollisionTest(pt.x, pt.y, rect);	}

inline bool CollisionTest(RECT r1, RECT r2)
{
	return CollisionTest(r1.left,  r1.top, r2) || CollisionTest(r1.left,  r1.bottom, r2) ||
		   CollisionTest(r1.right, r1.top, r2) || CollisionTest(r1.right, r1.bottom, r2) ||
		   CollisionTest(r2.left,  r2.top, r1) || CollisionTest(r2.left,  r2.bottom, r1) ||
		   CollisionTest(r2.right, r2.top, r1) || CollisionTest(r2.right, r2.bottom, r1);
}

inline bool CollisionTest(int r1left, int r1top, int r1right, int r1bottom,
				   int r2left, int r2top, int r2right, int r2bottom)
{
	return CollisionTest(r1left,  r1top,    r2left, r2top, r2right, r2bottom) || 
		   CollisionTest(r1left,  r1bottom, r2left, r2top, r2right, r2bottom) ||
		   CollisionTest(r1right, r1top,    r2left, r2top, r2right, r2bottom) || 
		   CollisionTest(r1right, r1bottom, r2left, r2top, r2right, r2bottom) ||
		   CollisionTest(r2left,  r2top,    r1left, r1top, r1right, r1bottom) || 
		   CollisionTest(r2left,  r2bottom, r1left, r1top, r1right, r1bottom) ||
		   CollisionTest(r2right, r2top,    r1left, r1top, r1right, r1bottom) || 
		   CollisionTest(r2right, r2bottom, r1left, r1top, r1right, r1bottom);
}

// BALL ROUTINES
void InitBalls()
{
	for(int i=0; i<BALL_COUNT; i++)
	{
		balls[i].x  = RANDOM(8, SCREEN_WIDTH-8);
		balls[i].y  = RANDOM(8, SCREEN_HEIGHT-8);
		balls[i].dx = RANDOM(-4,4);
		balls[i].dy = RANDOM(0,6);
	}
}

void ProcessBalls()
{
	for(int i=0; i<BALL_COUNT; i++)
	{
		balls[i].x	+= balls[i].dx;
		balls[i].y	+= balls[i].dy;				

		if( balls[i].x<0 || (balls[i].x+BALL_SIZE)>SCREEN_WIDTH )
		{
			balls[i].dx = -balls[i].dx;
			balls[i].x += balls[i].dx;
		}
		if( balls[i].y<0 || (balls[i].y+BALL_SIZE)>SCREEN_HEIGHT )
		{
			balls[i].dy = -balls[i].dy;
			balls[i].y += balls[i].dy;
		}

		// test collision with other balls
		for(int j=i; j<BALL_COUNT; j++)
		{
			RECT r1, r2;
			r1.left   = balls[i].x;
			r1.top    = balls[i].y;
			r1.right  = balls[i].x + BALL_SIZE;
			r1.bottom = balls[i].y + BALL_SIZE;

			r2.left   = balls[j].x;
			r2.top    = balls[j].y;
			r2.right  = balls[j].x + BALL_SIZE;
			r2.bottom = balls[j].y + BALL_SIZE;

			if( CollisionTest(r1,r2) )
			{
				balls[i].dx = -balls[i].dx;
				balls[i].dy = -balls[i].dy;
				balls[j].dx = -balls[j].dx;
				balls[j].dy = -balls[j].dy;
			}
		}
	}
}

void DrawBalls()
{
	for(int i=0; i<BALL_COUNT; i++)
		Draw_Rectangle(balls[i].x, balls[i].y, balls[i].x+BALL_SIZE, balls[i].y+BALL_SIZE, 255);
}


int Game_Main(void *parms)
{
	if(game_state==GAME_STATE_INIT)
	{
		srand(Start_Clock());	// seed random-number generator

		InitBalls();

		game_state=GAME_STATE_START_LEVEL;
	}
	else if(game_state==GAME_STATE_START_LEVEL)
	{
		game_state=GAME_STATE_RUN;
	}
	else if(game_state==GAME_STATE_RUN)
	{
		// start timing clock
		Start_Clock();

		Draw_Rectangle(0,0,SCREEN_WIDTH-1,SCREEN_HEIGHT-1,5); //clear screen

		ProcessBalls();
		DrawBalls();
		
		if(KEY_DOWN(VK_ESCAPE))	
			PostQuitMessage(0);

		// sync
		Wait_Clock(30);

		DD_Flip();
	}

	return 1;
} // end Game_Main

///////////////////////////////////////////////////////////////